package org.kohsuke.rng2srng;

import java.io.OutputStream;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.xml.transform.Result;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.sax.SAXTransformerFactory;
import javax.xml.transform.sax.TransformerHandler;

import org.kohsuke.rng2srng.datatype.DatatypeImpl;
import org.relaxng.datatype.Datatype;
import org.xml.sax.Attributes;
import org.xml.sax.ContentHandler;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.AttributesImpl;
import org.xml.sax.helpers.LocatorImpl;

import com.thaiopensource.relaxng.impl.EmptyPattern;
import com.thaiopensource.relaxng.impl.Name;
import com.thaiopensource.relaxng.impl.NameClass;
import com.thaiopensource.relaxng.impl.NameClassVisitor;
import com.thaiopensource.relaxng.impl.Pattern;
import com.thaiopensource.relaxng.impl.PatternVisitor;

/**
 * Walks the pattern tree and produces the pattern
 * in the simplified RELAX NG syntax as SAX events.
 * 
 * @author Kohsuke Kawaguchi (kk@kohsuke.org)
 */
public class Writer implements PatternVisitor, NameClassVisitor {
    
    /**
     * Sends the simplified RELAX NG format to the specified handler.
     */
    public static void write( Pattern p, ContentHandler receiver ) throws SAXException {
        receiver.startDocument();
        receiver.setDocumentLocator(new LocatorImpl());
        receiver.startPrefixMapping("",NSURI);
        new Writer(receiver).doWrite(p);
        receiver.endPrefixMapping("");
        receiver.endDocument();
    }
    
    public static void write( Pattern p, Result receiver ) throws SAXException {
        try {
            // use XSLT to change the format of the output
            TransformerHandler noopTransformer = 
                ((SAXTransformerFactory)TransformerFactory.newInstance()).newTransformerHandler();
            noopTransformer.setResult(receiver);
            
            write( p, noopTransformer );
        } catch( TransformerConfigurationException e ) {
            throw new SAXException(e);
        }
    }
    
    
    
    private void doWrite( Pattern p ) throws SAXException {
        try {
            start("grammar","xmlns",NSURI);
            start("start");
            p.accept(this);
            end("start");
            
            while(!queue.isEmpty()) {
                ElementDecl e = (ElementDecl)queue.remove(0);
    
                start("define", "name", (String)visitedElements.get(e) );
                start("element");
                e.nameClass.accept(this);
                e.contentModel.accept(this);
                end("element");
                end("define");
            }
            
            end("grammar");
        } catch( WrappedSAXException e ) {
            throw e.nested;
        }
    }
    
    
    private Writer( ContentHandler receiver ) {
        setReceiver(receiver);
    }
    
    /** This object will receive SAX events. */
    private ContentHandler receiver;
    
    public final void setReceiver( ContentHandler h ) {
        this.receiver = h;
    }
    
    /** The empty Attributes object. Used as a constant object. */
    private final static Attributes emptyAttributes = new AttributesImpl();
    
    /** RELAX NG namespace URI. */
    private final static String NSURI = "http://relaxng.org/ns/structure/1.0";
    
    /** SAXException wrapped as a RuntimException. */
    private class WrappedSAXException extends RuntimeException {
        WrappedSAXException( SAXException e ) { this.nested=e; }
        public final SAXException nested;
    }
    
    /** Produces a start tag. */
    private void start( String tagName ) {
        try {
            receiver.startElement(NSURI,tagName,tagName,emptyAttributes);
        } catch( SAXException e ) {
            throw new WrappedSAXException(e);
        }
    }
    /** Produces a start tag. */
    private void start( String tagName, String att1Name, String att1Value ) {
        start( tagName, new String[]{att1Name,att1Value});
    }
    
    private void start( String tagName, String[] attPairs ) {
        try {
            AttributesImpl atts = new AttributesImpl();
            
            for( int i=0; i<attPairs.length; i+=2 )
                atts.addAttribute("",attPairs[i],attPairs[i],"CDATA",attPairs[i+1]);
                
            receiver.startElement(NSURI,tagName,tagName,atts);
        } catch( SAXException e ) {
            throw new WrappedSAXException(e);
        }
    }
    
    /** Produces an end tag. */
    private void end( String tagName ) {
        try {
            receiver.endElement(NSURI,tagName,tagName);
        } catch( SAXException e ) {
            throw new WrappedSAXException(e);
        }
    }
    
    private void single( String tagName ) {
        start(tagName);
        end(tagName);
    }
    
    private void text( String text ) {
        try {
            receiver.characters(text.toCharArray(),0,text.length());
        } catch( SAXException e ) {
            throw new WrappedSAXException(e);
        }
    }
    


//
// Pattern visitor implementation
//
    public void visitAttribute(NameClass nc, Pattern p) {
        start("attribute");
        nc.accept(this);
        p.accept(this);
        end("attribute");
    }

    public void visitChoice(Pattern p1, Pattern p2) {
        if(p2 instanceof EmptyPattern)
            visitBinary("choice",p2,p1);
        else
            visitBinary("choice",p1,p2);
    }
    
    private void visitBinary( String tagName, Pattern p1, Pattern p2 ) {
        start(tagName);
        p1.accept(this);
        p2.accept(this);
        end(tagName);
    }

    public void visitData(Datatype dt) {
        startData( (DatatypeImpl)dt );
        end("data");
    }

    public void visitDataExcept(Datatype dt, Pattern except) {
        startData( (DatatypeImpl)dt );
        start("except");
        except.accept(this);
        end("except");
        end("data");
    }
    
    private void startData( DatatypeImpl dt ) {
        start("data", new String[]{"datatypeLibrary",dt.nsURI,"type",dt.name});
        for( int i=0; i<dt.parameters.size(); i++ ) {
            String[] p = (String[])dt.parameters.get(i);
            start("param", "name",p[0]);
            text(p[1]);
            end("param");
        }
    }

    public void visitValue(Datatype _dt, Object v) {
        DatatypeImpl dt = (DatatypeImpl)_dt;
        start("value", new String[]{"datatypeLibrary",dt.nsURI,"type",dt.name,"ns","TODO"});
        text(v.toString());
        end("value");
    }


    /** Set of visited patterns. A map from ElementDecl to name. */
    private final Map visitedElements = new HashMap();
    private int iota=1;
    
    /** List of unprocessed element decls. */
    private final List queue = new LinkedList();
    
    public void visitElement(NameClass nc, Pattern p) {
        ElementDecl e = new ElementDecl(nc,p);
        
        String name = (String)visitedElements.get(e);
        if( name==null ) {
            // this is the first time we saw this element decl.
            name = "element"+(iota++);
            visitedElements.put(e,name);
            queue.add(e);
        }
        
        start("ref", "name",name);
        end("ref");
    }

    public void visitEmpty() {
        single("empty");
    }

    public void visitError() {
        single("error");
    }

    public void visitGroup(Pattern p1, Pattern p2) {
        visitBinary("group",p1,p2);
    }

    public void visitInterleave(Pattern p1, Pattern p2) {
        visitBinary("interleave",p1,p2);
    }

    public void visitList(Pattern p) {
        start("list");
        p.accept(this);
        end("list");
    }

    public void visitNotAllowed() {
        single("notAllowed");
    }

    public void visitOneOrMore(Pattern p) {
        start("oneOrMore");
        p.accept(this);
        end("oneOrMore");
    }

    public void visitText() {
        single("text");
    }
    
    public void visitAnyName() {
        single("anyName");
    }

    public void visitAnyNameExcept(NameClass except) {
        start("anyName");
        start("except");
        except.accept(this);
        end("except");
        end("anyName");
    }

    public void visitChoice(NameClass nc1, NameClass nc2) {
        start("choice");
        nc1.accept(this);
        nc2.accept(this);
        end("choice");
    }

    public void visitName(Name name) {
        start("name", "ns", name.getNamespaceUri());
        text(name.getLocalName());
        end("name");
    }

    public void visitNsName(String ns) {
        start("nsName", "ns", ns);
        end("nsName");
    }

    public void visitNsNameExcept(String ns, NameClass except) {
        start("nsName", "ns", ns);
        start("except");
        except.accept(this);
        end("except");
        end("nsName");
    }

    public void visitNull() {
        single("null");
    }



    private class ElementDecl {
        ElementDecl( NameClass nameClass, Pattern contentModel ) {
            this.nameClass = nameClass;
            this.contentModel = contentModel;
        }
        
        private final NameClass nameClass;
        private final Pattern   contentModel;
        
        public boolean equals( Object o ) {
            if(!(o instanceof ElementDecl))     return false;
            ElementDecl rhs = (ElementDecl)o;
            
            return nameClass.equals(rhs.nameClass) && contentModel.equals(rhs.contentModel);
        }
        
        public int hashCode() {
            return nameClass.hashCode() ^ contentModel.hashCode();
        }
    }
}
